import scipy.special
import numpy as np
import random
import math
import Global as g

def log(*a):
    # print(*a)
    pass

class Layer():

    def __init__(self, name, num_rows, num_cols):
        log('Create', name, 'layer')
        self.name = name
        self.num_rows = num_rows
        self.num_cols = num_cols
        self.size = num_rows * num_cols  # number of neurons is this layer
        self.w = None          # weight matrix
        self.previous = None   # reference to layer below (if any)
        self.excitation = None # vector

    def set_(self, _pattern):
        self.excitation = _pattern

    def get_excitation(self):              
        return self.excitation

    def forward(self):
        _sum = np.dot(self.w, self.previous.get_excitation())
        self.excitation = _sum / self.previous.size

    def connect(self, previous):
        self.previous = previous
        self.w = np.random.normal(loc=0.5, scale=0.1, size=
                                  (self.size, self.previous.size))    
              
    def adjust_weight(self, row, col, win_row, win_col, input_vector, delta):
        index = row * self.num_cols + col
        # adjust weight
        diff = input_vector - self.w[index]   
        self.w[index] += diff * delta
        # Calculate error
        if row  == win_row and col == win_col:
            len_diff = np.linalg.norm(diff)
            len_input = np.linalg.norm(input_vector)
            return len_diff / len_input # error
        else:
            return 0
        
    def get_weights(self, row, col):
        w_as_matrix = self.w.reshape(self.num_rows, self.num_cols, self.previous.size)

#         print('Kohonen layer weights:')
#         for r in range(self.num_rows):
#             for c in range(self.num_cols):
#                 print('Row', r, 'Col', c, end=' ')
#                 for w in range(self.previous.size):
#                     val = w_as_matrix[r][c][w]
#                     if val < 0.01:
#                         print('....', end=' ')
#                     else:
#                         print(f'{val:0.2f}', end=' ')
#                 print()
                    
        vector = w_as_matrix[row][col]
        return vector
    
    def to_dict(self):
        rounded_weights = np.round(self.w, decimals=4)
        data = {'weights': rounded_weights.tolist()}
        return data

class KohonenNet():

    def __init__(self, num_input_rows, num_input_cols,
                 num_kohonen_rows, num_kohonen_cols, random_seed):
        log("KohonenNet.initialize: num_input_rows =", num_input_rows,
            ", num_input_cols =", num_input_cols,
            ", num_kohonen_rows =", num_kohonen_rows,
            ", num_kohonen_cols =", num_kohonen_cols,
            ", random_seed =", random_seed)

        random.seed(random_seed)
        np.random.seed(random_seed)
        self.input_layer = Layer("input layer", num_input_rows, num_input_cols)
        self.kohonen_layer = Layer("kohonen layer", num_kohonen_rows, num_kohonen_cols)
        self.kohonen_layer.connect(self.input_layer)

    def train(self, epoch, all_pattern, learning_rate, radius):
        log('KohonenNet.train: epoch =', epoch, 'number of pattern =', len(all_pattern),
            ', learning_rate =', learning_rate, ', radius =', radius)

        error = 0
        for pattern in all_pattern:
            # Select and apply training pattern
            output_index, input_vector = pattern # output_index = expected label
            self.input_layer.set_(input_vector)
            self.kohonen_layer.forward()
            # Find winning neuron (BMU)
            koh_excit = self.kohonen_layer.get_excitation()
            number_rows = number_cols = self.kohonen_layer.num_rows
            all_excit = koh_excit.reshape(number_rows, number_cols)
            max_excitation = -100
            for row in range(number_rows):
                for col in range(number_cols):
                    neuron_excit = all_excit[row][col]
                    if neuron_excit > max_excitation:
                        max_excitation = neuron_excit
                        winning_neuron = (row, col)
            if epoch != None:
                g.history.append((epoch, output_index, winning_neuron))
            # Apply learning step

            win_row, win_col = winning_neuron
            for row in range(number_rows):
                for col in range(number_cols):            
                    # calculate hat function
                    dist2 = (row - win_row)**2 + (col - win_col)**2
                    # The original function:
                    hat = math.exp(-dist2 / radius**2)
                    # End original function
                    # Mexican hat start
#                     r2 = radius**2
#                     hat = (1 - dist2 / r2) * math.exp(-dist2 / r2 / 2)
                    # Mexican hat end
                    delta = hat * learning_rate
                    # adapt neuron
                    error += self.kohonen_layer.adjust_weight(row, col, win_row, win_col, input_vector, delta)            
        return error    

    def run(self, pattern_data):
        self.input_layer.set_(pattern_data)
        self.kohonen_layer.forward()
        kohonen_layer_excitation = self.kohonen_layer.get_excitation()
        return kohonen_layer_excitation

    def get_kohonen_weights(self, row, column):
        return self.kohonen_layer.get_weights(row, column)      
        
    def dump(self):
        print('Dumping trained network')
        return {'kohonen_layer': self.kohonen_layer.to_dict()}
