from copy import deepcopy
import tkinter as tk
import numpy as np
import scipy.special
import Global as g

class Map(tk.Frame):

    def excitation_color(self, value, min_value, max_value):
        if self.autoranging.get():
            denominator = (max_value - min_value)
            if denominator == 0:
                value = 0
            else:
                value = (value - min_value) / denominator
        
        if self.radio_kohonen_display.get() == 'heatmap':
            rgb = self.heat_map(value)
        else:
            bright = 255 - int(value * 255)
            rgb = bright, bright, bright
        
        return rgb
        
#     def grayscale(self, value, max_value):
#         max_value = max(0.01, max_value) # avoid overflow if max_value == 0
#         if self.autoranging.get():
#             value = min(1.0, value / max_value)
#         bright = 255 - int(value * 255)
#         
#         return bright, bright, bright
    
    def heat_map(self, value):
        r = [0,     0,  65, 255, 255, 255, 252]
        g = [0,   195, 255, 255, 150,  18,   0]
        b = [255, 255,   0,   0,   0,   7, 255]

        interval = 1 / (len(r) - 1)
        whole, rem = divmod(value, interval)
        if whole >= len(r) - 1:
            return r[-1], g[-1], b[-1]

        whole = int(whole)
        frac = rem / interval
        red   = int(r[whole] + frac * (r[whole + 1] - r[whole]))
        green = int(g[whole] + frac * (g[whole + 1] - g[whole]))
        blue  = int(b[whole] + frac * (b[whole + 1] - b[whole]))  

        return red, green, blue

    def create_manual_pattern(self):
        # print(self.radio_input_mode.get())
        self.button_clear.configure(state=tk.NORMAL)
        self.spin_box_record.configure(state=tk.DISABLED)
        index = int(self.spin_box_record.get()) - 1
        if len(self.pattern) == 0:
            return
        label, data = self.pattern[index]
        self.input_layer_excitation = deepcopy(data).reshape(self.input_layer_rows, self.input_layer_cols)
        self.show_manual_pattern()

    def show_manual_pattern(self):
        for row in range(self.input_layer_rows):
            for col in range(self.input_layer_cols):
                bright = 255 - int(self.input_layer_excitation[row][col] * 255)
                color = "#%02x%02x%02x" % (bright, bright, bright)
                tag = self.input_layer_squares[row][col]
                self.canvas_input.itemconfig(tag, fill=color)

    def clear_manual_pattern(self):
        self.input_layer_excitation = np.zeros((self.input_layer_rows, self.input_layer_cols))
        self.show_manual_pattern()
        self.kohonen_excitation = g.kohonen_net.run(self.input_layer_excitation.flatten())
        self.show_kohonen_excitation()

    def show_input_pattern(self, index):          
        if len(self.pattern) == 0:
            return
        self.spin_box_record.configure(state=tk.NORMAL)
        self.button_clear.configure(state=tk.DISABLED)
        self.spin_box_record.configure(from_=1, to=len(self.pattern))
        self.current_record.set(index+1)
        pattern, data = self.pattern[index]
        self.pattern_label.set(self.pattern_labels[pattern])
        data = data.reshape(self.input_layer_rows, self.input_layer_cols)
        for row in range(self.input_layer_rows):
            for col in range(self.input_layer_cols):
                bright = 255 - int(data[row][col] * 255)
                color = "#%02x%02x%02x" % (bright, bright, bright)
                tag = self.input_layer_squares[row][col]
                self.canvas_input.itemconfig(tag, fill=color)
                
    def show_kohonen_excitation(self):
        number_rows = number_cols = g.kohonen_layer_size
        all_excit = self.kohonen_excitation.reshape(number_rows, number_cols)
        
#         print('Kohonen layer excitation:')
#         for r in range(number_rows):
#             print('Row', r, end=' ')
#             for c in range(number_cols):
#                 val = all_excit[r][c]
#                 if val < 0.01:
#                     print('....', end=' ')
#                 else:
#                     print(f'{val:0.3f}', end=' ')
#             print()

        min_excitation = 100
        max_excitation = -100
        for row in range(number_rows):
            for col in range(number_cols):
                neuron_excit = all_excit[row][col]
                if neuron_excit > max_excitation:
                    max_excitation = neuron_excit
                    self.winning_neuron = (row, col)
                if neuron_excit < min_excitation:
                    min_excitation = neuron_excit
                
        for row in range(number_rows):
            for col in range(number_cols):
                neuron_excit = all_excit[row][col]
                rgb = self.excitation_color(neuron_excit, min_excitation, max_excitation)
                color = "#%02x%02x%02x" % rgb
                tag = self.kohonen_layer_squares[row][col]
                self.canvas_kohonen.itemconfig(tag, fill=color)
                
        if self.show_winning_neuron.get():
            win_row, win_col = self.winning_neuron
            tag = self.kohonen_layer_squares[win_row][win_col]
            if self.radio_kohonen_display.get() == 'heatmap' :
                color = 'black'
            else:
                color = 'red'
            self.canvas_kohonen.itemconfig(tag, fill=color)
            self.current_excitation.set(f'{max_excitation:0.4f}')
            self.show_synaptic_weights(win_row, win_col)
            
 
    def spin_record(self):
        index = self.spin_box_record.get()
        self.apply_pattern(int(index) - 1)

    def create_display(self):
        self.input_layer_rows = int(g.num_input_layer_rows.get())
        self.input_layer_cols = int(g.num_input_layer_cols.get())
        
        self.create_input_display()
        self.create_kohonen_display()
        self.create_weights_display()

    def create_input_display(self):
        
        self.canvas_input.delete(tk.ALL)
        self.input_layer_squares = [[None for c in range(self.input_layer_cols)]
                                          for r in range(self.input_layer_rows)]
        
        canvas_width = int(self.canvas_input.cget('width'))
        canvas_height = int(self.canvas_input.cget('height'))
        self.input_neuron_size = min(canvas_height // self.input_layer_rows,
                               canvas_width // self.input_layer_cols)
        if self.input_neuron_size > 40: self.input_neuron_size = 40
        for row in range(self.input_layer_rows):
            for col in range(self.input_layer_cols):
                x0 = self.input_neuron_size * col
                x1 = x0 + self.input_neuron_size
                y0 = self.input_neuron_size * row
                y1 = y0 + self.input_neuron_size
                handle = self.canvas_input.create_rectangle(x0, y0, x1, y1,
                                                           fill=g.bgLight,
                                                           outline="#808080")
                self.input_layer_squares[row][col] = handle

    def create_kohonen_display(self):
        self.canvas_kohonen.delete(tk.ALL)
        number_rows = number_cols = g.kohonen_layer_size

        self.kohonen_layer_squares = [[None for c in range(number_cols)]
                                            for r in range(number_rows)]
        
        canvas_width = int(self.canvas_kohonen.cget('width'))
        canvas_height = int(self.canvas_kohonen.cget('height'))
        self.kohonen_neuron_size = min(canvas_height // number_rows,
                               canvas_width // number_cols)
        if self.kohonen_neuron_size > 40: self.kohonen_neuron_size = 40
        for row in range(number_rows):
            for col in range(number_cols):
                x0 = self.kohonen_neuron_size * col
                x1 = x0 + self.kohonen_neuron_size
                y0 = self.kohonen_neuron_size * row
                y1 = y0 + self.kohonen_neuron_size
                handle = self.canvas_kohonen.create_rectangle(x0, y0, x1, y1,
                                                           fill=g.bgLight,
                                                           outline="#808080")
                self.kohonen_layer_squares[row][col] = handle

    def create_weights_display(self):
        self.canvas_weights.delete(tk.ALL)
        self.weight_squares = [[None for c in range(self.input_layer_cols)]
                               for r in range(self.input_layer_rows)]
        
        canvas_width = int(self.canvas_weights.cget('width'))
        canvas_height = int(self.canvas_weights.cget('height'))
        self.input_neuron_size = min(canvas_height // self.input_layer_rows,
                               canvas_width // self.input_layer_cols)
        if self.input_neuron_size > 40: self.input_neuron_size = 40
        for row in range(self.input_layer_rows):
            for col in range(self.input_layer_cols):
                x0 = self.input_neuron_size * col
                x1 = x0 + self.input_neuron_size
                y0 = self.input_neuron_size * row
                y1 = y0 + self.input_neuron_size
                handle = self.canvas_weights.create_rectangle(x0, y0, x1, y1,
                                                           fill=g.bgLight,
                                                           outline="#808080")
                self.weight_squares[row][col] = handle

    def apply_pattern(self, index):
        self.radio_input_mode.set('pattern')
        self.button_clear.configure(state=tk.DISABLED)
        self.spin_box_record.configure(state=tk.NORMAL)
        
        if len(g.all_test_patterns) > 0:
            self.pattern = g.all_test_patterns
            self.pattern_labels = g.test_pattern_labels
        else:
            self.pattern = g.all_training_patterns
            self.pattern_labels = g.training_pattern_labels
            
        if index < 0:
            index = int(self.spin_box_record.get()) - 1
        # Check index validity
        if len(self.pattern) - index < 1:
            return
        # Show the input pattern
        self.show_input_pattern(index)
        
        # Run the network
        if not g.kohonen_net:
            return
        self.output_index, data = self.pattern[index]
        self.kohonen_excitation = g.kohonen_net.run(data)
        self.show_kohonen_excitation()

    def spin_box_record_changed(self, _):
        val = self.spin_box_record.get()
        self.current_record.set(int(val))
        self.apply_pattern(int(val)-1)

    def input_canvas_mouse_down(self, event):
        if self.radio_input_mode.get() != "manual":
            return
        # print('Click at x=', event.x, ', y=', event.y)
        col = int(event.x // self.input_neuron_size)
        row = int(event.y // self.input_neuron_size)

        if row < self.input_layer_rows and col < self.input_layer_cols:
            # print("Input layer clicked at", row, col)
            handle = self.input_layer_squares[row][col]
            # print('Create copy of input_layer_excitation')
            self.modified_input_layer_excitation = deepcopy(self.input_layer_excitation)
            self.input_layer_was_copied = True
            if self.input_layer_excitation[row][col] < 0.5:
                self.modified_input_layer_excitation[row][col] = 1.0
                self.canvas_input.itemconfig(handle, fill="black")
            else:
                self.modified_input_layer_excitation[row][col] = 0.0
                self.canvas_input.itemconfig(handle, fill="white")
        else:
            pass
            # print("Missed canvas")

    def input_canvas_drag(self, event):
        if self.radio_input_mode.get() != "manual":
            return
        # print("Mouse dragging over canvas:", event.x, ":", event.y)
        if not self.input_layer_was_copied:
            return

        col = int(event.x // self.input_neuron_size)
        row = int(event.y // self.input_neuron_size)
        if row < self.input_layer_rows and col < self.input_layer_cols:
            # print("Input layer dragged at", row, col)
            handle = self.input_layer_squares[row][col]
            self.modified_input_layer_excitation[row][col] = 1.0
            self.canvas_input.itemconfig(handle, fill="black")

    def input_canvas_mouse_up(self, _):
        pass
        if self.radio_input_mode.get() != "manual":
            return
        if self.input_layer_was_copied:
            # print('Replace input_layer_excitation')
            self.input_layer_excitation = self.modified_input_layer_excitation
            self.input_layer_was_copied = False
            self.kohonen_excitation = g.kohonen_net.run(self.input_layer_excitation.flatten())
            self.show_kohonen_excitation()

    def show_synaptic_weights(self, row, col):
        weights_vector = g.kohonen_net.get_kohonen_weights(row, col)
        # print('Raw weights', weights_vector)  ## print
        weights = weights_vector.reshape(self.input_layer_rows, self.input_layer_cols)
        for row in range(self.input_layer_rows):
            for col in range(self.input_layer_cols):
                bright = 255 - int(weights[row][col] * 255)
                # print(bright, end=' ')
                color = "#%02x%02x%02x" % (bright, bright, bright)
                tag = self.weight_squares[row][col]
                self.canvas_weights.itemconfig(tag, fill=color)
            # print()
            
    def kohonen_mouse_hover(self, event):
        if not self.kohonen_neuron_size:
            return
        row = event.y // self.kohonen_neuron_size
        if row >= g.kohonen_layer_size: return
        col = event.x // self.kohonen_neuron_size
        if col >= g.kohonen_layer_size: return
        all_excit = self.kohonen_excitation.reshape(g.kohonen_layer_size,
                                                    g.kohonen_layer_size)
        neuron_excit = all_excit[row][col]
        self.current_excitation.set(f'{neuron_excit:0.4f}')
        self.show_synaptic_weights(row, col)
    
    def mousewheel_record(self, event):
        if event.delta > 0:
            self.spin_box_record.invoke('buttonup')
        else:
            self.spin_box_record.invoke('buttondown')
        
    column_1 = 15
    column_2 = column_1 + 100
    column_3 = column_2 + 150
    lineSpace = 40
    header_line = 10
    row_1 = 46
    row_2 = 270
    
    def __init__(self, notebook):
        super(Map, self).__init__(notebook, background=g.bgDark)
        
        self.pattern = []
        self.kohonen_neuron_size = None
        self.winning_neuron = (0, 0)
        self.rows = [self.row_2 + n * self.lineSpace for n in range(30)]

        # Left Pane
        frame = tk.Frame(self, height=550, width=200, background=g.bgDark,
                         borderwidth=2, relief=tk.SUNKEN)
        frame.place(x=10, y=10)
        
        lbl = tk.Label(frame, text='Input Layer', font=g.fontTitle, background=g.bgDark)
        lbl.place(x=self.column_1, y=self.header_line)

        self.canvas_input = tk.Canvas(frame, width=155, height=370,
                                      bg=g.bgDark, highlightbackground=g.bgDark)
        self.canvas_input.place(x=self.column_1, y=self.row_1)
        self.canvas_input.bind('<B1-Motion>', self.input_canvas_drag)
        self.canvas_input.bind('<Button-1>', self.input_canvas_mouse_down)
        self.canvas_input.bind('<ButtonRelease-1>', self.input_canvas_mouse_up)
        self.input_layer_squares = []
        
        lbl = tk.Label(frame, text='Label', font=g.fontLabel, background=g.bgDark)
        self.pattern_label = tk.StringVar()
        lbl.place(x=self.column_1, y=self.rows[0])
        lbl = tk.Label(frame, textvariable=self.pattern_label,
                                     font=g.fontLabel, width=6, background=g.bgLight)
        lbl.place(x=self.column_2, y=self.rows[0])

        # Radio buttons
        self.radio_input_mode = tk.StringVar()
        self.input_layer_excitation = None
        self.modified_input_layer_excitation = None
        self.input_layer_was_copied = False
        self.radio_button = tk.Radiobutton(frame, text="Manual", variable=self.radio_input_mode, value='manual',
                                        command=self.create_manual_pattern, font=g.fontLabel, background=g.bgDark)
        self.radio_button.place(x=self.column_1, y=self.rows[1])

        self.button_clear = tk.Button(frame, text="Clear", width=6, state=tk.DISABLED,
                                      font=g.fontLabel, background=g.bgDark, command=self.clear_manual_pattern)
        self.button_clear.place(x=self.column_2, y=self.rows[1])

        self.radio_button = tk.Radiobutton(frame, text="Record #", variable=self.radio_input_mode, value='pattern',
                                        command=lambda: self.apply_pattern(-1), font=g.fontLabel, background=g.bgDark)
        self.radio_button.place(x=self.column_1, y=self.rows[2])
        self.radio_button.select()


        self.current_record = tk.IntVar()
        self.current_record.set(1)
        self.spin_box_record = tk.Spinbox(self, from_=1, to=1, textvariable=self.current_record,
                                          font=g.fontTitle, width=4, background=g.bgDark, command=self.spin_record)
        self.spin_box_record.place(x=self.column_2 + 10, y=self.rows[2] + 13)
        self.spin_box_record.bind('<Return>', self.spin_box_record_changed)
        self.spin_box_record.bind('<MouseWheel>', self.mousewheel_record)

        # Center Pane
        frame = tk.Frame(self, height=550, width=400, background=g.bgDark,
                         borderwidth=2, relief=tk.SUNKEN)
        frame.place(x=220, y=10)

        lbl = tk.Label(frame, text='Kohonen Layer', font=g.fontTitle, background=g.bgDark)
        lbl.place(x=self.column_1, y=self.header_line)

        self.canvas_kohonen = tk.Canvas(frame, width=360, height=360,
                                      bg=g.bgDark, highlightbackground=g.bgDark)
        self.canvas_kohonen.place(x=self.column_1, y=self.row_1)
        self.canvas_kohonen.bind('<Motion>', self.kohonen_mouse_hover)
        self.kohonen_layer_squares = []
        
        self.show_winning_neuron = tk.IntVar()
        self.show_winning_neuron.set(1)
        chk = tk.Checkbutton(frame, text='Show winning neuron', font=g.fontLabel,
                             bg=g.bgDark, variable=self.show_winning_neuron,
                             command=self.show_kohonen_excitation)
        
        chk.place(x=self.column_1, y=self.rows[4])

        self.radio_kohonen_display = tk.StringVar()
        self.radio_button = tk.Radiobutton(frame, text="Grayscale", variable=self.radio_kohonen_display, value='grayscale',
                                        command=self.show_kohonen_excitation, font=g.fontLabel, background=g.bgDark)
        self.radio_button.place(x=self.column_1, y=self.rows[5])
        self.radio_button.select()
        
        self.radio_button = tk.Radiobutton(frame, text="Heatmap", variable=self.radio_kohonen_display, value='heatmap',
                                        command=self.show_kohonen_excitation, font=g.fontLabel, background=g.bgDark)
        self.radio_button.place(x=self.column_1, y=self.rows[6])

        self.autoranging = tk.IntVar()
        self.autoranging.set(0)
        chk = tk.Checkbutton(frame, text='Autoranging', font=g.fontLabel,
                             bg=g.bgDark, variable=self.autoranging,
                             command=self.show_kohonen_excitation)       
        chk.place(x=self.column_2+20, y=self.rows[6])

        self.current_excitation = tk.StringVar()
        self.current_excitation.set('')
        lbl = tk.Label(frame, textvariable=self.current_excitation,
                        font=g.fontLabel, width=6, background=g.bgLight)
        lbl.place(x=self.column_3, y=self.rows[4])
        lbl = tk.Label(frame, text='Excitation of\nselected neuron', font=g.fontLabel,
                       justify=tk.LEFT, background=g.bgDark)
        lbl.place(x=self.column_3, y=self.rows[4]+25)
        
        # Right Pane
        frame = tk.Frame(self, height=550, width=200, background=g.bgDark,
                         borderwidth=2, relief=tk.SUNKEN)
        frame.place(x=630, y=10)

        lbl = tk.Label(frame, text='Synaptic Weights', font=g.fontTitle, background=g.bgDark)
        lbl.place(x=self.column_1, y=self.header_line)

        self.canvas_weights = tk.Canvas(frame, width=155, height=370,
                                      bg=g.bgDark, highlightbackground=g.bgDark)
        self.canvas_weights.place(x=self.column_1, y=self.row_1)
        self.weight_squares = []

        lbl = tk.Label(frame, text='of the selected neuron\nin Kohonen Layer',
                       justify=tk.LEFT, font=g.fontLabel, background=g.bgDark)
        lbl.place(x=self.column_1, y=self.rows[0])
        
