# NeuroLab port to Python with extensions
# Martin Reiche 2024
import tkinter as tk
from tkinter import ttk
from SetTrainTest import SetTrainTest
from Map import Map
from History import History
import Global as g

class Gui(tk.Tk):
    def __init__(self):
        super(Gui, self).__init__()
        self.title('NeuroLab 2.0 - martin-reiche.de 2024')
        self.minsize(842, 685)
        self.resizable(0, 0)

        # Tabulator
        customed_style = ttk.Style()
        customed_style.theme_use("default")
        customed_style.configure('TNotebook.Tab', padding=[6, 6], font=g.fontLabel)
        customed_style.map('TNotebook.Tab', background=[('selected', g.bgDark)])
        customed_style.configure('TButton', foreground='white')

        tab_control = ttk.Notebook(self, style='Custom.TNotebook')
        tab_control.bind('<<NotebookTabChanged>>', self.tab_changed)
        
        # Setup, Train & Test tab
        g.tab_setup_train_test = SetTrainTest(tab_control)
        tab_control.add(g.tab_setup_train_test, text='Setup, Train & Test')
        tab_control.pack(expand=1, fill='both')

        # Map tab
        g.tab_kohonen_map = Map(tab_control)
        tab_control.add(g.tab_kohonen_map, text='Kohonen Map')
        tab_control.pack(expand=1, fill='both')
        
        # History tab
        g.tab_history = History(tab_control)
        tab_control.add(g.tab_history, text='Training History')
        tab_control.pack(expand=1, fill='both')        
        
        # Status bar
        g.test_file_name = tk.StringVar()
        status_bar_1 = tk.Label(self, textvariable=g.test_file_name, bd=1, relief=tk.SUNKEN, anchor=tk.W,
                                font=("Arial", 10), pady=4, background=g.bgDark)
        status_bar_1.pack(side=tk.BOTTOM, fill=tk.X)

        g.training_file_name = tk.StringVar()
        status_bar_2 = tk.Label(self, textvariable=g.training_file_name, bd=1, relief=tk.SUNKEN, anchor=tk.W,
                                font=("Arial", 10), pady=4, background=g.bgDark)
        status_bar_2.pack(side=tk.BOTTOM, fill=tk.X)

        self.current_status = tk.StringVar()
        self.status_bar_3 = tk.Label(self, textvariable=self.current_status, bd=1, relief=tk.SUNKEN, anchor=tk.W,
                                font=("Arial", 12), pady=4, background=g.bgDark, fg='yellow')
        self.status_bar_3.pack(side=tk.BOTTOM, fill=tk.X)

    def update_status(self, text):
        self.current_status.set(text)
        self.status_bar_3.update()
        
    def tab_changed(self,event):
        if event.widget.index('current') == 1: # map tab
            g.tab_kohonen_map.apply_pattern(-1)
        elif  event.widget.index('current') == 2: # history tab
            g.tab_history.prepare_history()
            
g.gui = Gui() # The gui is a global instance!
g.gui.mainloop()

